#include <math.h>
#include <stdlib.h>

#include "fib.h"

void
swap(int *a, int *b)
{
	*a ^= *b;
	*b ^= *a;
	*a ^= *b;
}

struct fib_heap *
fh_makeheap()
{
	struct fib_heap *n;

	if ((n = malloc(sizeof *n)) == NULL)
		return NULL;

	fh_initheap(n);

	return n;
}

void
fh_initheap(struct fib_heap *new)
{
	new->fh_cmp_fnct = NULL;
	new->fh_neginf = NULL;
	new->fh_n = 0;
	new->fh_min = NULL;
	new->fh_root = NULL;
}

void
fh_setcmp(struct fib_heap *h, void_cmp_fnct fnct)
{
	h->fh_cmp_fnct = fnct;
}

void
fh_setneginf(struct fib_heap *h, void *data)
{
	h->fh_neginf = data;
}

/*
 * this will return these values:
 *	NULL	failed for some reason
 *	ptr	token to use for manipulation of data
 */
void *
fh_insert(struct fib_heap *h, void *data)
{
	struct fib_heap_el *x;

	if ((x = fhe_newelem()) == NULL)
		return NULL;

	fhe_setdata(x, data);
	fh_insertrootlist(h, x);

	if (h->fh_min == NULL ||
	    h->fh_cmp_fnct(x->fhe_data, h->fh_min->fhe_data) < 0)
		h->fh_min = x;
	h->fh_n++;

	return (void *)x;
}

void *
fh_min(struct fib_heap *h)
{
	if (h->fh_min == NULL)
		return NULL;
	return h->fh_min->fhe_data;
}

/* this needs to be implemented:
struct fib_heap *fh_union(struct fib_heap *, struct fib_heap *);*/

void *
fh_extractmin(struct fib_heap *h)
{
	struct fib_heap_el *z;
	void *a = NULL;

	z = h->fh_min;
	if (z != NULL) {
		struct fib_heap_el *x, *y, *orig;

		a = z->fhe_data;
		orig = NULL;
		for(x = z->fhe_child; x != orig && x != NULL;) {
			if (orig == NULL)
				orig = x;
			y = x->fhe_right;
			x->fhe_p = NULL;
			fh_insertrootlist(h, x);
			x = y;
		}
		fh_removerootlist(h, z);
		h->fh_n--;
		if (z == z->fhe_right)
			h->fh_min = NULL;
		else {
			h->fh_min = z->fhe_right;
			fh_consolidate(h);
		}
		free(z);
	}
	return a;
}

void *
fh_replacekey(struct fib_heap *h, void *a, void *k)
{
	struct fib_heap_el *x = (struct fib_heap_el *)a;
	struct fib_heap_el *y;

	if (h->fh_cmp_fnct(k, x->fhe_data) > 0)
		return NULL;
	x->fhe_data = k;
	y = x->fhe_p;
	if (y != NULL && h->fh_cmp_fnct(x->fhe_data, y->fhe_data) < 0) {
		fh_cut(h, x, y);
		fh_cascading_cut(h, y);
	}
	if (h->fh_cmp_fnct(x->fhe_data, h->fh_min->fhe_data) < 0)
		h->fh_min = x;

	return x;
}

void
fh_delete(struct fib_heap *h, void *a)
{
	struct fib_heap_el *x = (struct fib_heap_el *)a;

	fh_replacekey(h, a, h->fh_neginf);
	fh_extractmin(h);
}

/*
 * begin of private fib_heap fuctions
 */
void
fh_insertrootlist(struct fib_heap *h, struct fib_heap_el *x)
{
	if (h->fh_root == NULL) {
		h->fh_root = x;
		x->fhe_left = x;
		x->fhe_right = x;
		return;
	}

	fhe_insertafter(h->fh_root, x);
}

void
fh_removerootlist(struct fib_heap *h, struct fib_heap_el *x)
{
	if (x->fhe_left == x)
		h->fh_root = NULL;
	else {
		fhe_remove(x);
		if (h->fh_root == x)
			h->fh_root = x->fhe_right;
	}
}

void
fh_consolidate(struct fib_heap *h)
{
	struct fib_heap_el **a;
	struct fib_heap_el *w;
	struct fib_heap_el *y;
	struct fib_heap_el *x;
	int i;
	int d;
	int D;

	D = (int) ceil(log(h->fh_n)/log(2))+1;
	a = (struct fib_heap_el **)malloc(sizeof *a * D);
	for (i = 0; i < D; i++)
		a[i] = NULL;
	while ((w = h->fh_root) != NULL) {
		x = w;
		fh_removerootlist(h, w);
		x->fhe_left = x;
		x->fhe_right = x;
		d = x->fhe_degree;
		while(a[d] != NULL) {
			y = a[d];
			if (h->fh_cmp_fnct(x->fhe_data, y->fhe_data) > 0)
				swap((int *)&x, (int *)&y);
			fh_heaplink(h, y, x);
			a[d] = NULL;
			d++;
		}
		a[d] = x;
	}
	h->fh_min = NULL;
	for (i =0; i < D; i++)
		if (a[i] != NULL) {
			fh_insertrootlist(h, a[i]);
			if (h->fh_min == NULL || h->fh_cmp_fnct(a[i]->fhe_data,
			    h->fh_min->fhe_data) < 0)
				h->fh_min = a[i];
		}
}

void
fh_heaplink(struct fib_heap *h, struct fib_heap_el *y, struct fib_heap_el *x)
{
	/* make y a child of x */
	if (x->fhe_child == NULL)
		x->fhe_child = y;
	else
		fhe_insertbefore(x->fhe_child, y);
	y->fhe_p = x;
	x->fhe_degree++;
	y->fhe_mark = 0;
}

void
fh_cut(struct fib_heap *h, struct fib_heap_el *x, struct fib_heap_el *y)
{
	fhe_remove(x);
	y->fhe_degree--;
	fh_insertrootlist(h, x);
	x->fhe_p = NULL;
	x->fhe_mark = 0;
}

void
fh_cascading_cut(struct fib_heap *h, struct fib_heap_el *y)
{
	struct fib_heap_el *z;

	z = y->fhe_p;
	if (z != NULL)
		if (y->fhe_mark == 0)
			y->fhe_mark = 1;
		else {
			fh_cut(h, y, z);
			fh_cascading_cut(h, z);
		}
}

/*
 * begining of handling elements of fib_heap
 */
struct fib_heap_el *
fhe_newelem()
{
	struct fib_heap_el *e;

	if ((e = malloc(sizeof *e)) == NULL)
		return NULL;

	fhe_initelem(e);

	return e;
}

void
fhe_initelem(struct fib_heap_el *e)
{
	e->fhe_degree = 0;
	e->fhe_mark = 0;
	e->fhe_p = NULL;
	e->fhe_child = NULL;
	e->fhe_left = e;
	e->fhe_right = e;
	e->fhe_data = NULL;
}

void
fhe_setdata(struct fib_heap_el *e, void *data)
{
	e->fhe_data = data;
}

void
fhe_insertafter(struct fib_heap_el *a, struct fib_heap_el *b)
{
	if (a == a->fhe_right) {
		a->fhe_right = b;
		a->fhe_left = b;
		b->fhe_right = a;
		b->fhe_left = a;
	} else {
		b->fhe_right = a->fhe_right;
		a->fhe_right->fhe_left = b;
		a->fhe_right = b;
		b->fhe_left = a;
	}
}

void
fhe_insertbefore(struct fib_heap_el *a, struct fib_heap_el *b)
{
	fhe_insertafter(a->fhe_left, b);
}

void
fhe_remove(struct fib_heap_el *x)
{
	x->fhe_right->fhe_left = x->fhe_left;
	x->fhe_left->fhe_right = x->fhe_right;
}
